/*-------------------------------------------------------------------------
 *
 * recoverycmds.c
 *	  commands for backup, restore, and PITR
 *
 * Copyright (c) 2002, PostgreSQL Global Development Group
 *
 *
 * IDENTIFICATION
 *	  $Header: /home/jnield/cvswork/pgsql-server/src/backend/commands/recoverycmds.c,v 1.1 2002/08/21 12:38:11 jnield Exp $
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include "access/heapam.h"
#include "access/xlogdefs.h"
#include "access/xlog.h"
#include "access/clog.h"
#include "catalog/catname.h"
#include "catalog/catalog.h"
#include "catalog/pg_database.h"
#include "catalog/pg_class.h"

#include "catalog/indexing.h"
#include "commands/comment.h"
#include "commands/recoverycmds.h"
#include "lib/stringinfo.h"
#include "miscadmin.h"
#include "nodes/parsenodes.h"
#include "parser/parser.h"
#include "storage/freespace.h"
#include "storage/sinval.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/fmgroids.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

#ifdef MULTIBYTE
#include "mb/pg_wchar.h"		/* encoding check */
#endif

/* XXX: TODO: Work with Win32 port people to fix this */
#ifndef DIRSEPERATOR
#define DIRSEPERATOR '/'
#endif

/* TODO: These will become GUC vars */
static mode_t		RCCBackupDirCreateMode 	= 0770;
static mode_t		RCCBackupFileCreateMode	= 0660;
static int			RCCBackupCopyBlocksPerWrite = 16;

static struct {
	time_t		bds_start;
	time_t		bds_end;
#ifndef INT64_IS_BUSTED
	uint64		bds_nblocks;
#else
	uint32		bds_nblocks;
#endif
	uint32		bds_nfiles;
	int			bds_logfiles_copied;
} bkStats;

#ifndef INT64_IS_BUSTED
#define BKStatAddBlocks( nblks ) (bkStats.bds_nblocks += (nblks))
#else
#define BKStatAddBlocks( nblks ) \
( \
 	(((nblks) + bkStats.bds_nblocks) - bkStats.bds_nblocks != nblks)? \
		bkStats.bds_nblocks = 0xFFFFFFFF: bkStats.bds_nblocks += (nblks) \
)
#endif

#define BKStatAddFiles( nfiles ) (bkStats.bds_nfiles += (nfiles))

static void PrintBackupStatistics(const char *targ_dpath);
static bool IsValidTargetDirectory(const char *dirpath);
static bool CreateTargetDirectory(const char *dirpath);
static bool IsValidBackupDirectory(const char *dirpath);
static char * DatabasePathBuild(const char *datadir, Oid tblNode);
static bool DatabaseParentDirsInstall(const char *datadir);
static bool CopyDatabaseFile(char *destpath, char *srcpath, bool allow_partial);
static bool CopyFileBlockAligned(char *destpath, char *srcpath, bool allow_partial);
static bool CopyDirRegularFiles(const char *destdpath, const char *srcdpath);
static bool CopyDatabaseDirectory(const char *dest_datadir, Oid dbid);
static List * GetDatabaseOidList(void);

/* TODO: put in header */
/* recent chane due to UseNewLine being declared static in postgres.c */

#ifndef TCOP_DONTUSENEWLINE
static int UseNewLine = 1;
#else
static int UseNewLine = 0;
#endif

void
AlterSystemBackup(AlterSystemBackupStmt *stmt)
{
	XLogRecPtr		lsnstart, lsnend;
	char			backup_ctlfile[MAXPGPATH+1];
	char			backup_ctltmp[MAXPGPATH];
	char			backup_logdir[MAXPGPATH];
	char			backup_clogdir[MAXPGPATH];
	char			tmppath[MAXPGPATH];
	List 			*dboids = NIL, *l;
	bool			in_backup = false;
	char			dest_dirpath[MAXPGPATH];

	if ( IsTransactionBlock() )
		elog(ERROR, 
			"ALTER SYSTEM BACKUP: may not be called in a transaction block");

	if ( !superuser() )
		elog(ERROR,
			"ALTER SYSTEM BACKUP: permission denied");

	if ( !stmt )
		elog(ERROR, "AlterSystemBackup: NULL statement (Parse Bug)");

	if ( !stmt->to_dest_dir )
	{
		if ( !stmt->force_next_logfile )
			elog(ERROR, "AlterSystemBackup: Invalid statment (Parse Bug)");

		elog(NOTICE, 
			"ALTER SYSTEM BACKUP FORCE NEXT: forcing advance to next log segment");

		if ( !XLogBeginBackup(NULL) ) /* Don't copy controlfile */
			elog(ERROR, "ALTER SYSTEM BACKUP FORCE NEXT: command failed");

		XLogCompleteBackup(&lsnstart, &lsnend);

		CommitTransactionCommand();
		/* All our memory context is nuked.  */
		StartTransactionCommand();
		(void)CreateCheckPointReturningLSN(false, true);

		CommitTransactionCommand();
		StartTransactionCommand();
		(void)CreateCheckPointReturningLSN(false, true);

		elog(NOTICE, "ALTER SYSTEM BACKUP FORCE NEXT: complete");
		return;
	}

	/* Init stats */
	time(&bkStats.bds_start);
	time(&bkStats.bds_end);
	bkStats.bds_nblocks = 0;
	bkStats.bds_nfiles = 0;
	bkStats.bds_logfiles_copied = 0;

	/* We will kill QueryContext later... */
	strncpy(dest_dirpath, stmt->to_dest_dir, MAXPGPATH);

	/* PART 1:
	 *   Ensure the destination directory is OK, and
	 *   setup XLOG for backup
	 */
	if ( !IsValidBackupDirectory(dest_dirpath) )
		elog(ERROR, "ALTER SYSTEM BACKUP: target directory '%s' is not useable",
				dest_dirpath);

	backup_ctltmp[MAXPGPATH] = '\0';
	snprintf(backup_ctltmp, MAXPGPATH, "%s/temporary_control_file.bak",
				dest_dirpath);

	if ( !XLogBeginBackup(backup_ctltmp) )
		elog(ERROR, "ALTER SYSTEM BACKUP: command cancelled");

	in_backup = true;

	closeAllVfds(); /* opendir() uses FD's */

	XLOGPathGenerate(backup_logdir, MAXPGPATH, backup_ctlfile, MAXPGPATH,
						dest_dirpath);

	/* PART 2:
	 *  Back up the files
	 */
	/* copy all the toplevel files */
	if ( !CopyDirRegularFiles(dest_dirpath, DataDir) )
	{
		elog(WARNING, "ALTER SYSTEM BACKUP: error copying toplevel files");
		goto L_errexit;
	}

	/* Make the immediate subdirs of the root */
	if ( !DatabaseParentDirsInstall(dest_dirpath) )
		goto L_errexit;
	
	/* Copy all the database directories */
	if ( (dboids = GetDatabaseOidList()) == NIL )
	{
		elog(WARNING,
				"ALTER SYSTEM BACKUP: failed to get database list");
		goto L_errexit;
	}
	/* Add the global "database" (Oid == 0) */
	dboids = lcons( (void *) ((Oid) 0), dboids );

	foreach( l, dboids )
	{
		if ( !CopyDatabaseDirectory(dest_dirpath, (Oid) lfirst(l)) )
		{
			elog(WARNING, 
					"ALTER SYSTEM BACKUP: error backing up database Oid %u",
					(uint32) lfirst(l));
			goto L_errexit;
		}
		elog(NOTICE, "copied database Oid %u", (uint32) lfirst(l));
	}

	/* Copy the CLOG */
	CLOGBuildPath(backup_clogdir, MAXPGPATH, dest_dirpath);
	if ( !CreateTargetDirectory(backup_clogdir) ||
			!CopyDirRegularFiles(backup_clogdir, GetCLOGDir()) )
	{
		elog(WARNING, "ALTER SYSTEM BACKUP: error copying CLOG files");
		goto L_errexit;
	}
	
	/* PART 3:
	 *   Finish with XLOG, and copy all the logfiles
	 *   we need to recover the fuzzy backup we just made.
	 */
	XLogCompleteBackup(&lsnstart, &lsnend);
	in_backup = false;

	/* Move the temporary backup controlfile into position
	 * as the new control file for the backup database.
	 */
	unlink(backup_ctlfile);
	if ( !CopyDatabaseFile(backup_ctlfile, backup_ctltmp, true) )
	{
		elog(WARNING, 
				"ALTER SYSTEM BACKUP: can't copy backup control file to %s",
				backup_ctlfile);
		goto L_errexit;
	}
	unlink(backup_ctltmp);

	/* Remove the postgresql.pid lock file from the backup database */
	snprintf(tmppath, MAXPGPATH, "%s%c%s", 
				dest_dirpath, DIRSEPERATOR, "postmaster.pid");
	if ( unlink(tmppath) != 0 )
	{
		elog(WARNING,
				"ALTER SYSTEM BACKUP: can't remove backup lock file '%s'",
				tmppath);
	}

	/* XLogCompleteBackup has made sure that the insert
	 * point has advanced to a new logfile beyond the one
	 * containing 'lsnend'. Now force it into the archive.
	 *
	 * Checkpoint TWICE to force our log into the archive.
	 *
	 * XXX: FIXME: We shouldn't have to do this twice. 
	 * CreateCheckPoint needs to be changed.
	 */
	CommitTransactionCommand();
	/* All our memory context is nuked.
	 * Watch out for this...
	 */
	dboids = NIL;

	StartTransactionCommand();
	(void)CreateCheckPointReturningLSN(false, true);

	CommitTransactionCommand();
	StartTransactionCommand();
	(void)CreateCheckPointReturningLSN(false, true);

	/* PART 4:
	 *   Copy the reqired logfiles into the destination directory.
	 */
	{
		char path[MAXPGPATH+1];

		path[MAXPGPATH] = '\0';
		snprintf(path, MAXPGPATH, "%s%c%s", 
					dest_dirpath, DIRSEPERATOR, "pg_xlog");
		
		/* XLogCopyArchive... will rename the files from the archive
		 * name to the normal name when 'rename_online' is true.
		 */
		bkStats.bds_logfiles_copied = 
			XLogCopyArchiveFileRangeContaining(path, true, lsnstart, lsnend);
		if ( bkStats.bds_logfiles_copied < 0 )
		{
			elog(WARNING, "ALTER SYSTEM BACKUP: failed to copy archived logs");
			goto L_errexit;
		}
		else if ( bkStats.bds_logfiles_copied == 0 )
		{
			elog(WARNING, 
					"No log files copied. Report BUG: %X/%X, %X/%X, %s",
					lsnstart.xlogid, lsnstart.xrecoff, 
					lsnend.xlogid, lsnend.xrecoff, path);
			goto L_errexit;
		}
	}
	time(&bkStats.bds_end);

	PrintBackupStatistics(dest_dirpath);

	return;

	/* NEVER REACH HERE */

L_errexit:
	if ( in_backup )
		{ XLogAbortBackup(); }
	elog(ERROR, "ALTER SYSTEM BACKUP: backup command aborted due to errors");
	return;
}

/*
 * Prompts the user for a command string, parses the string,
 * and stores an InteractiveRecoveryCmd into 'cmdout' describing
 * the action to take.
 */
void
InteractiveRecoveryPrompt( InteractiveRecoveryCmd cmdout )
{
	int				c;
	bool			end = false;
	bool			backslashSeen = false;
	StringInfo		inBuf;
	List			*parsetrees;
	List			*l;
	Node			*parseTree;
	AlterSystemRecoverStmt *rstmt = NULL;

	if ( IsUnderPostmaster )
		elog(ERROR, "InteractiveRecoveryPrompt: must be standalone backend");

	printf("recovery> ");
	fflush(stdout);

	inBuf = makeStringInfo();

	/* Adapted from InteractiveBackend() */
	for (;;)
	{
		if (UseNewLine)
		{
			/*
			 * if we are using \n as a delimiter, then read characters
			 * until the \n.
			 */
			while ((c = getc(stdin)) != EOF)
			{
				if (c == '\n')
				{
					if (backslashSeen)
					{
						/* discard backslash from inBuf */
						inBuf->data[--inBuf->len] = '\0';
						backslashSeen = false;
						continue;
					}
					else
					{
						/* keep the newline character */
						appendStringInfoChar(inBuf, '\n');
						break;
					}
				}
				else if (c == '\\')
					backslashSeen = true;
				else
					backslashSeen = false;

				appendStringInfoChar(inBuf, (char) c);
			}

			if (c == EOF)
				end = true;
		}
		else
		{
			/*
			 * otherwise read characters until EOF.
			 */
			while ((c = getc(stdin)) != EOF)
				appendStringInfoChar(inBuf, (char) c);

			if (inBuf->len == 0)
				end = true;
		}

		if (end)
			goto L_end;

		/*
		 * otherwise we have a user query so process it.
		 */
		break;
	}

	fflush(stdout);

	/* parse the query */
	{
		int old_eloglev = yyerrorElogLevel;

		yyerrorElogLevel = WARNING;
		parsetrees = raw_parser(inBuf->data);
		yyerrorElogLevel = old_eloglev;
	}

	/* Make sure there is only one parse tree resulting from that string */
	parseTree = NULL;
	foreach(l, parsetrees)
	{
		if ( !parseTree )
			{ parseTree = (Node *)lfirst(l); }
		else
		{
			elog(NOTICE, "ERROR: multiple commands in statement");
			goto L_empty_command;
		}
	}

	if ( !parseTree )
		{ goto L_empty_command; }

	/* Only care about a couple utility commands */
	switch ( nodeTag(parseTree) )
	{
		case T_AlterSystemRecoverStmt:
			rstmt = (AlterSystemRecoverStmt *)parseTree;

			if ( !rstmt->command )
			{
				elog(WARNING, "PARSE BUG: unexpected NULL in rstmt->command");
				goto L_empty_command;
			}

			if ( !strcasecmp("commit", rstmt->command) )
			{
				cmdout->irc_type = IRCT_FINISH;
			}
			else if ( !strcasecmp("from", rstmt->command) ||
						!strcasecmp("", rstmt->command) )
			{
				cmdout->irc_type = IRCT_RECOVER;
				cmdout->irc.reco.rc_flags = 0;
				if ( !rstmt->to_end )
				{
					cmdout->irc.reco.rc_flags |= IRC_RC_TO_POINT;
					cmdout->irc.reco.finish_point.xlogid = rstmt->xlogid;
					cmdout->irc.reco.finish_point.xrecoff = rstmt->xrecoff;
				}

				if ( rstmt->fname )
				{
					struct stat st;

					if ( stat(rstmt->fname, &st) != 0 )
					{
						elog(NOTICE, "directory '%s' is not valid: %m",
								rstmt->fname);
						goto L_empty_command;
					}
					if ( !S_ISDIR(st.st_mode) )
					{
						elog(NOTICE, "path '%s' is not a directory",
								rstmt->fname);
						goto L_empty_command;
					}
					if ( access(rstmt->fname, R_OK|X_OK) != 0 )
					{
						elog(NOTICE, "directory '%s' is not accessable",
								rstmt->fname);
						goto L_empty_command;
					}

					cmdout->irc.reco.rc_flags |= IRC_RC_FROM_SOURCE;
					/* XXX: alloc? */
					cmdout->irc.reco.source_dir = rstmt->fname;
				}
			}
			else
			{
				elog(NOTICE, "ALTER SYSTEM RECOVER: invalid command");
				goto L_empty_command;
			}
			break;

		default:
			elog(NOTICE, "Command not avalable during interactive recovery");
			goto L_empty_command;
	}

	return;

L_empty_command:
	cmdout->irc_type = IRCT_EMPTY;
	return;
	
L_end:
	cmdout->irc_type = IRCT_EOF;
	return;
}

static void
PrintBackupStatistics(const char *targ_dpath)
{
	elog(NOTICE,
			"ALTER SYSTEM BACKUP complete in %.2f minutes\n"
			"\tBACKUP was started at %s\n"
			"\tTarget directory was %s\n"
			"\t%u Data Files Copied\n"
			"\t%.0f Data Blocks Copied (Size %.2f MB)\n"
			"\t%d Log Files Copied from Archive",
			(double) (difftime(bkStats.bds_end, bkStats.bds_start)/60.0),
			ctime(&bkStats.bds_start), /* Don't use this twice... */
			targ_dpath,
			bkStats.bds_nfiles,
			(double)bkStats.bds_nblocks, 
			((double)bkStats.bds_nblocks)/
				(1024.0 * 1024.0 / (double)BLCKSZ),
			bkStats.bds_logfiles_copied);
}

static bool 
IsValidTargetDirectory(const char *dirpath)
{
	struct stat st;
	
	if ( stat(dirpath, &st) < 0 )
	{
		elog(WARNING, "Backup target %s: %m", dirpath);
		return false;
	}

	if ( !S_ISDIR(st.st_mode) )
	{
		elog(WARNING, "Backup target %s is not a directory", dirpath);
		return false;
	}

	if ( access(dirpath, R_OK|W_OK|X_OK) < 0 )
	{
		elog(WARNING, "Backup target %s: %m", dirpath);
		return false;
	}

	return true;
}

/* 
 * Create a directory sutable for use as a backup target if it
 * does not exist.
 *
 * RETURNS: 
 *   'true' if the target directory is valid and ready for use
 *   'false' otherwise
 */
static bool
CreateTargetDirectory(const char *dirpath)
{
	if ( mkdir(dirpath, RCCBackupDirCreateMode) < 0 )
	{
		if ( errno == EEXIST )
		{
			if ( IsValidTargetDirectory(dirpath) )
			{
				return true;
			}
			else
			{
				elog(WARNING, "Refusing to overwrite non-directory %s",
						dirpath);
				return false;
			}
		}
		else
		{
			elog(WARNING, "Can't create directory %s: %m", dirpath);
			return false;
		}
	}

	/* This might happen with inherited ACL's on some systems */
	if ( !IsValidTargetDirectory(dirpath) )
	{
		elog(WARNING, 
			"Directory %s is not useable by me, even though I just created it",
			dirpath);
		return false;
	}
	return true;
}

/*
 * Is this directory OK for use as a backup target?
 *
 * RETURNS: 'true' if it can be used
 * 			'false' otherwise
 */
static bool
IsValidBackupDirectory(const char *dirpath)
{
	DIR 		*d;
	struct dirent *dent;

	if ( !IsValidTargetDirectory(dirpath) )
		return false;

	if ( !(d = opendir(dirpath)) )
	{
		elog(WARNING, "Can't open backup directory %s: %m", dirpath);
		return false;
	}

	while ( (dent = readdir(d)) )
	{
		if ( strcmp(dent->d_name, ".") &&
			 strcmp(dent->d_name, "..") )
		{
			/* This is allowed, but warn the user */
			elog(WARNING, "Backup directory %s is not empty", dirpath);
			break;
		}
	}

	closedir(d);
	return true;
}

/*
 * Return a path to the database with Oid 'tblNode' given
 * a database rooted at 'datadir'
 */
static char *
DatabasePathBuild(const char *datadir, Oid tblNode)
{
	char	*path;

	if ( tblNode == (Oid) 0)
	{
		path = (char *) palloc(strlen(datadir) + 8);
		sprintf(path, "%s%cglobal", datadir, DIRSEPERATOR);
	}
	else
	{
		path = (char *) palloc(strlen(datadir) + 6 + sizeof(NameData) + 1);
		sprintf(path, "%s%cbase%c%u", 
				datadir, DIRSEPERATOR, DIRSEPERATOR, tblNode);
	}
	return path;
}

/*
 * Create the basic toplevel directories for a database.
 *
 * RETURNS: 'true' on success, 'false' on failure
 */
static bool
DatabaseParentDirsInstall(const char *datadir)
{
	char	path[MAXPGPATH+1];
	bool	result = true;

	path[MAXPGPATH] = '\0';

	snprintf(path, MAXPGPATH, "%s%cglobal", datadir, DIRSEPERATOR);
	result = CreateTargetDirectory(path) && result;

	snprintf(path, MAXPGPATH, "%s%cbase", datadir, DIRSEPERATOR);
	result = CreateTargetDirectory(path) && result;

	snprintf(path, MAXPGPATH, "%s%cpg_xlog", datadir, DIRSEPERATOR);
	result = CreateTargetDirectory(path) && result;

	snprintf(path, MAXPGPATH, "%s%cpg_arch_dir", datadir, DIRSEPERATOR);
	result = CreateTargetDirectory(path) && result;

	return result;
}
static bool
CopyDatabaseFile(char *destpath, char *srcpath, bool allow_partial)
{
	return CopyFileBlockAligned(destpath, srcpath, allow_partial);
}

#if 0
static bool
CopyStaticFile(char *destpath, char *srcpath, bool allow_partial)
{
	return CopyFileBlockAligned(destpath, srcpath, allow_partial);
}

static bool
CopyFileNodeUsingBufferManager(char *destpath, Oid srcTblNode, Oid srcRelNode)
{
	RelationData	pseudoRel;
	BufferDesc		*bufHdr;
	LockRelId		relid;

	pseudoRel.rd_fd = -1;
	pseudoRel.rd_node.tblNode = srcTblNode;
	pseudoRel.rd_node.relNode = srcRelNode;

	relid.tblNode = srcTblNode;
	relid.relNode = srcRelNode;

	{

		LWLockAcquire(BufMgrLock, LW_EXCLUSIVE);
		/* XXX: We rely on BufferAlloc not to care that
	 	* our relation isn't real. All it should use are the
	 	* values set above.
	 	*/
		bufHdr = BufferAlloc(&pseudoRel, i, &found);
	}

}
#endif

/*
 * Our read request size must be exactly the same as the
 * buffer manager, and must be aligned on block boundaries.
 *
 * If allow_partial is true, no error will be reported if the file
 * does not end on a block boundary.
 */
static bool
CopyFileBlockAligned(char *destpath, char *srcpath, bool allow_partial)
{
	static char * buf = NULL;
	static size_t bufsz = 0;
	static int    bufblks = 0;
	int			  srcrcode, destrcode;
	size_t		  final_sz;
	File		  srcf=0, destf=0;
	int			  i;

	Assert( strcmp(srcpath, destpath) );

	/* Allocate once first time.
	 */
	if ( !buf )
	{
		bufblks = RCCBackupCopyBlocksPerWrite;
		bufsz   = BLCKSZ * bufblks;
		if ( !(buf = malloc(bufsz)) )
			elog(ERROR, "Out of memory");
	}

	
	if ( (srcf = FileNameOpenFile(srcpath, O_RDONLY | PG_BINARY, 0)) < 0 )
	{
		elog(LOG, "Unable to open file %s: %m", srcpath);
		return false;
	}

	if ( (destf = FileNameOpenFile(destpath, O_CREAT|O_EXCL|O_RDWR|PG_BINARY, 
									  RCCBackupFileCreateMode)) < 0 )
	{
		elog(WARNING, "Unable to create file %s: %m", destpath);
		return false;
	}

	
	for ( i=0;
		  (srcrcode = FileRead(srcf, &buf[BLCKSZ*i], BLCKSZ)) == BLCKSZ;
		  i = (i + 1) % bufblks )
	{
		if ( i+1 == bufblks &&
			 (destrcode = FileWrite(destf, buf, bufsz)) != bufsz )
		{
			if ( destrcode < 0 )
			{
				elog(WARNING, "Error writing %s: %m", destpath);
				goto L_cleanup;
			}
			else
			{
				elog(WARNING, "Partial write to %s (requested %dB got %dB)",
						destpath, bufsz, destrcode);
				goto L_cleanup;
			}
		}
		BKStatAddBlocks(1);
	}

	if ( srcrcode < 0 )
	{
		elog(WARNING, "Error reading %s: %m", srcpath);
		goto L_cleanup;
	}
	else if ( srcrcode > 0 && !allow_partial )
	{
		elog(WARNING, "Partial block read from %s", srcpath);
		goto L_cleanup;
	}

	final_sz = BLCKSZ*i + srcrcode;

	if (final_sz > 0 && 
		 (destrcode = FileWrite(destf, buf, final_sz)) != final_sz)
	{
		if ( destrcode < 0 )
		{
			elog(WARNING, "Error writing %s: %m", destpath);
			goto L_cleanup;
		}
		else
		{
			elog(WARNING, "Incomplete write to %s (requested %dB got %dB)",
					destpath, final_sz, destrcode);
			goto L_cleanup;
		}
	}
	BKStatAddBlocks(i);

	FileClose(srcf);
	FileClose(destf);
	BKStatAddFiles(1);
	return true;

L_cleanup:
	if ( srcf )
		FileClose(srcf);
	if ( destf )
		FileClose(destf);
	return false;
}

/*
 * Copy all the regular files in a directory (non-recursive).
 */
static bool
CopyDirRegularFiles(const char *destdpath, const char *srcdpath)
{
	DIR  *dir = NULL;
	struct dirent *dent;
	char srcfname[MAXPGPATH+1], destfname[MAXPGPATH+1];
	char * srcbname;
	size_t srcbnsz;
	char * destbname;
	size_t destbnsz;
	struct stat st;

	srcfname[MAXPGPATH]  = '\0';
	destfname[MAXPGPATH] = '\0';

	snprintf(srcfname, MAXPGPATH, "%s%c", srcdpath, DIRSEPERATOR);
	srcbname = &srcfname[strlen(srcfname)];
	srcbnsz = MAXPGPATH - strlen(srcfname);

	snprintf(destfname, MAXPGPATH, "%s%c", destdpath, DIRSEPERATOR);
	destbname = &destfname[strlen(destfname)];
	destbnsz = MAXPGPATH - strlen(destfname);

	if ( !(dir = opendir(srcdpath)) )
	{
		elog(WARNING, "Can't open directory '%s': %m", srcdpath);
		goto L_cleanup;
	}

	while ( (dent = readdir(dir)) )
	{
		strncpy(srcbname, dent->d_name, srcbnsz);
		if ( strcmp(srcbname, dent->d_name) != 0 )
		{
			elog(WARNING, "Path overflow (%s)", dent->d_name);
			goto L_cleanup;
		}

		if ( stat(srcfname, &st) < 0 )
		{
			elog(WARNING, "Can't stat() file '%s': %m", srcfname);
			goto L_cleanup;
		}
		if ( !S_ISREG(st.st_mode) )
			{ continue; }

		strncpy(destbname, dent->d_name, destbnsz);
		if ( strcmp(destbname, dent->d_name) != 0 )
		{
			elog(WARNING, "Path overflow (%s)", dent->d_name);
			goto L_cleanup;
		}

		if ( !CopyDatabaseFile(destfname, srcfname, true) )
		{
			/* It is allowed for files to be deleted from
			 * under us.
			 */
			if ( stat(srcfname, &st) && 
					errno == ENOENT )
			{
				elog(LOG, "File %s was deleted during backup", srcfname);
				continue;
			}
			else
			{
				elog(WARNING, "Copy of file %s failed", srcfname);
				goto L_cleanup;
			}
		}
	}

	closedir(dir);
	return true;

L_cleanup:
	if ( dir )
		{ closedir(dir); }
	return false;
}

/*
 * RETURNS: 'true' on success, 'false' on failure.
 */
static bool
CopyDatabaseDirectory(const char *dest_datadir, Oid dbid)
{
	char *srcpath;
	char *destpath;
	bool result;

	srcpath = DatabasePathBuild(DataDir, dbid);
	destpath = DatabasePathBuild(dest_datadir, dbid);
	if ( !CreateTargetDirectory(destpath) )
		{ return false; }

	result = CopyDirRegularFiles(destpath, srcpath);
	pfree(srcpath);
	pfree(destpath);
	return result;
}

/*
 * Get a list of all the existing databases.
 *
 * RETURNS: NIL on failure.
 */
static List *
GetDatabaseOidList(void)
{
	List				*result = NIL;
	Datum				dboid;
	HeapTuple			tup;
	HeapScanDesc		scan;
	Relation			rel;

	Assert( sizeof(void *) >= sizeof(Oid) );

	rel = heap_openr(DatabaseRelationName, ShareLock);
	/* There must be at least one database, so this works for error cond. */
	if ( !RelationIsValid(rel) )
	{
		elog(WARNING, "heap_openr failed for %s", DatabaseRelationName);
		return NIL;
	}

	if ( !(scan = heap_beginscan(rel, SnapshotNow, 0, NULL)) )
	{
		elog(WARNING, "heap_beginscan failed for %s",
				RelationGetRelationName(rel));
		goto L_cleanup_open;
	}
	
	while ( (tup = heap_getnext(scan, ForwardScanDirection)) &&
			HeapTupleIsValid(tup) )
	{
		if ( !(dboid = heap_getsysattr(tup, ObjectIdAttributeNumber, NULL)) )
		{
			elog(WARNING, "heap_getsysattr failed");
			goto L_cleanup_scan;
		}
		result = lcons( (void *) DatumGetObjectId(dboid), result );
	}

L_cleanup_scan:
	heap_endscan(scan);
L_cleanup_open:
	heap_close(rel, NoLock);

	return result;
}
